#define QUOTE_NULL -1

struct QUOTE
{
  int LP;
  long TimeMsc;

  double bid;
  double ask;

  QUOTE( void ) : LP(QUOTE_NULL)//, bid(0)
  {
  }

  bool Set( const string Str )
  {
    string StrArray[];
    const bool Res = (::StringSplit(Str, '\t', StrArray) >= 4);

    if (Res)
    {
      this.TimeMsc = (long)StrArray[0];
      this.LP = (int)StrArray[1];

      this.bid = (double)StrArray[2];
      this.ask = (double)StrArray[3];
    }

    return(Res);
  }

  bool IsNull( void ) const
  {
    return(this.LP == QUOTE_NULL);
  }

  void ToNull( void )
  {
    this.LP = QUOTE_NULL;

    return;
  }

  int GetLiveTime( const long Time ) const
  {
    return((int)(Time - this.TimeMsc));
  }

  bool operator >( const QUOTE &Quote ) const
  {
    const double Price = ::MathMin(Quote.bid, Quote.ask);

    return((this.bid > Price) || ((this.bid == Price) && (this.TimeMsc > Quote.TimeMsc)));
  }

  bool operator <( const QUOTE &Quote ) const
  {
    const double Price = ::MathMax(Quote.bid, Quote.ask);

    return((this.ask < Price) || ((this.ask == Price) && (this.TimeMsc > Quote.TimeMsc)));
  }

#define TOSTRING(A) #A + " = " + (string)(A) + " "
  string ToString( void ) const
  {
    return(TOSTRING(LP) + TOSTRING(TimeMsc) + TOSTRING(bid) + TOSTRING(ask));
  }

  string ToStringBid( void ) const
  {
    return(TOSTRING(LP) + TOSTRING(TimeMsc) + TOSTRING(bid));
  }

  string ToStringAsk( void ) const
  {
    return(TOSTRING(LP) + TOSTRING(TimeMsc) + TOSTRING(ask));
  }
#undef TOSTRING
};

#undef QUOTE_NULL

class MARKET
{
private:
  const int MaxLiveTime;

  void CheckQuotesTime( const long Time )
  {
    for (int i = this.AmountLP - 1; i >= 0; i--)
      if (!this.Quotes[i].IsNull() && (this.Quotes[i].GetLiveTime(Time) > this.MaxLiveTime))
        this.Quotes[i].ToNull();

    return;
  }

protected:
  const int AmountLP;
  QUOTE Quotes[];

public:
  MARKET( const int iMaxLiveTime, const int iAmountLP ) : MaxLiveTime(iMaxLiveTime),
                                                          AmountLP(::ArrayResize(this.Quotes, iAmountLP))
  {
  }

  void operator +=( const QUOTE &Quote )
  {
    this.Quotes[Quote.LP] = Quote;

    this.CheckQuotesTime(Quote.TimeMsc);

    return;
  }

  string ToString( void ) const
  {
    string Str = NULL;

    for (int i = 0; i < this.AmountLP; i++)
      if (!this.Quotes[i].IsNull())
        Str += "\n" + this.Quotes[i].ToString();

    return(Str);
  }
};

class LEVEL2 : public MARKET
{
private:
  const double MaxSpread;
  QUOTE BestBand;

  bool Init;

  template <typename T>
  static void ArrayInsert( T &Array[], const int Pos, const T &Value )
  {
    for (int i = ::ArrayResize(Array, ::ArraySize(Array) + 1) - 1; i > Pos; i--)
      Array[i] = Array[i - 1];

    Array[Pos] = Value;

    return;
  }

  static void AddBid( const QUOTE &Quote, QUOTE &Bids[] )
  {
    const int Size = ::ArraySize(Bids);

    int Count = 0;

    while (Count < Size)
    {
      if (Quote > Bids[Count])
      {
        LEVEL2::ArrayInsert(Bids, Count, Quote);

        break;
      }

      Count++;
    }

    if (Count == Size)
      LEVEL2::ArrayInsert(Bids, Count, Quote);

    return;
  }

  static void AddAsk( const QUOTE &Quote, QUOTE &Asks[] )
  {
    const int Size = ::ArraySize(Asks);

    int Count = 0;

    while (Count < Size)
    {
      if (Quote < Asks[Count])
      {
        LEVEL2::ArrayInsert(Asks, Count, Quote);

        break;
      }

      Count++;
    }

    if (Count == Size)
      LEVEL2::ArrayInsert(Asks, Count, Quote);

    return;
  }

  void CheckSpread( QUOTE &Bids[], QUOTE &Asks[] )
  {
    if (::ArraySize(Bids) && ::ArraySize(Asks))
    {
      if (!this.Init)
      {
        this.BestBand.bid = (Asks[0].ask + Bids[0].bid) / 2;
        this.BestBand.ask = this.BestBand.bid;

        this.Init = true;
      }

      if (Asks[0].ask - Bids[0].bid > this.MaxSpread)
      {
        if (Bids[0] > this.BestBand)
        {
          LEVEL2::AddAsk(this.BestBand, Asks);

          if (Asks[0].ask < Bids[0].bid)
            Asks[0].ask = Bids[0].bid;
        }
        else if (Asks[0] < this.BestBand)
        {
          LEVEL2::AddBid(this.BestBand, Bids);

          if (Bids[0].bid > Asks[0].ask)
            Bids[0].bid = Asks[0].ask;
        }
        else
        {
          LEVEL2::AddBid(this.BestBand, Bids);
          LEVEL2::AddAsk(this.BestBand, Asks);
        }
      }

      this.BestBand.TimeMsc = ::MathMax(Bids[0].TimeMsc, Asks[0].TimeMsc);

      this.BestBand.bid = Bids[0].bid;
      this.BestBand.ask = Asks[0].ask;
    }

    return;
  }

public:
  LEVEL2( const double dMaxSpread, const int iMaxLiveTime, const int iAmountLP = 3 ) : MARKET(iMaxLiveTime, iAmountLP),
                                                                                       MaxSpread(dMaxSpread), Init(false)
  {
  }

  void Get( QUOTE &Bids[], QUOTE &Asks[] )
  {
    for (int i = this.AmountLP - 1; i >= 0; i--)
      if (!this.Quotes[i].IsNull())
      {
        LEVEL2::AddBid(this.Quotes[i], Bids);
        LEVEL2::AddAsk(this.Quotes[i], Asks);
      }

    this.CheckSpread(Bids, Asks);

    return;
  }

  string ToString( void )
  {
    string Str = "Market:" + this.MARKET::ToString();

    QUOTE Bids[];
    QUOTE Asks[];

    this.Get(Bids, Asks);

    Str += "\nLevel2:";

    for (int i = ::ArraySize(Asks) - 1; i >= 0; i--)
      Str += "\n" + Asks[i].ToStringAsk() + (i ? NULL : " *");

    const int Size = ArraySize(Bids);

    for (int i = 0; i < Size; i++)
      Str += "\n" + Bids[i].ToStringBid() + (i ? NULL : " *");

    return(Str);
  }
};

int GetQuotes(const string FileName, QUOTE &Quotes[] )
{
  uchar Bytes[];
  string Str[];

  const int Size = (FileLoad(FileName, Bytes) > 0) ? ArrayResize(Quotes, StringSplit(CharArrayToString(Bytes), '\r', Str) - 1) : 0;
  int Amount = 0;

  for (int i = 1; i < Size; i++)
    if (Quotes[Amount].Set(Str[i]))
      Amount++;

  return(ArrayResize(Quotes, Amount));
}

#property script_show_inputs

input string inFileNameIN = __FILE__ + "\\QuotesIN.txt"; //    
input string inFileNameOUT = __FILE__ + "\\QuotesOUT.txt"; //    

input group "Level2 settings"
input double inMaxSpread = 0.02; //  
input int inMaxTime = 5; //    
input int inAmountLP = 3; //  LP

void OnStart()
{
  const int handle = FileOpen(inFileNameOUT, FILE_WRITE | FILE_TXT);

  if (handle != INVALID_HANDLE)
  {
    QUOTE Quotes[];

    const int Size = GetQuotes(inFileNameIN, Quotes);

    LEVEL2 Level2(inMaxSpread, inMaxTime, inAmountLP);

    FileWriteString(handle, "MaxTime = " + (string)inMaxTime + ", MaxSpread = " + (string)inMaxSpread + "\n\n");

    for (int i = 0; i < Size; i++)
    {
      Level2 += Quotes[i];

      FileWriteString(handle, "NewTick:\n" + Quotes[i].ToString() + "\n\n" + Level2.ToString() + "\n---------------\n");
    }

    MessageBox("Output information - " + inFileNameOUT);

    FileClose(handle);
  }
}